__version__ = "0.1.4"

from pytorch_tao.core import ConfigMissingError, ensure_config, load_cfg
from pytorch_tao.repo import DirtyRepoError, Repo

try:
    repo = Repo.find_by_file(".")
    cfg = load_cfg(repo.cfg_path)
except FileNotFoundError:
    repo = None
    cfg = None

args = None

__all__ = [
    "Repo",
    "load_cfg",
    "cfg",
    "ensure_config",
    "ConfigMissingError",
    "args",
    "DirtyRepoError",
]



import os
import shutil
from copy import copy
from datetime import datetime
from pathlib import Path
from tempfile import mkdtemp
from typing import Union

import git
from torch.distributed.run import run

import pytorch_tao as tao
from pytorch_tao import core


class DirtyRepoError(Exception):
    """Raised if action need clean repo but it is not"""


class Repo:
    def __init__(self, path: Union[Path, str]):
        if isinstance(path, str):
            path = Path(path)
        self.path = path
        self.name = path.name
        self.tao_path = path / ".tao"
        self.git_path = path / ".git"
        self.cfg_path = self.tao_path / "cfg.yml"
        if self.path.exists():
            try:
                self.git = git.Repo(self.path)
            except git.InvalidGitRepositoryError:
                pass

    def init(self):
        self.tao_path.mkdir(exist_ok=False)
        config_content = core.render_tpl(
            "cfg.yml", name=self.name, run_dir=(self.tao_path / "runs").as_posix()
        )
        self.cfg_path.write_text(config_content)
        gitignore_content = core.render_tpl(".gitignore")
        (self.tao_path / ".gitignore").write_text(gitignore_content)

        self.git = git.Repo.init(self.path)
        self.git.git.add(all=True)
        self.git.index.commit("initial commit")

    def load_cfg(self):
        tao.load_cfg(self.cfg_path)

    def exists(self):
        """Is this tao repo exists and valid"""
        return self.path.exists() and self.tao_path.exists() and self.git_path.exists()

    @tao.ensure_config("run_dir")
    def run(self):
        """Start a training process.

        Run will call torch.distributed.run.run so this func will rely on the
        command line arguments. Call this function with right command line options.
        """
        if tao.args.tao_commit:
            self.git.git.add(all=True)
            self.git.index.commit(tao.args.tao_commit)
        if not tao.args.tao_dirty and self.git.is_dirty(untracked_files=True):
            raise DirtyRepoError()
        run_dir = Path(tao.cfg["run_dir"])
        run_dir = run_dir if run_dir.is_absolute() else self.path / run_dir
        metadata = {
            "dirty": self.git.is_dirty(untracked_files=True),
            "commit": self.git.head.ref.commit.hexsha,
            "run_at": datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
        }
        if tao.args.tao_dirty:
            # run process in current pwd
            run_fold = self.path
        else:
            # run process in run dir
            hexsha = self.git.head.ref.commit.hexsha[:8]
            run_fold = run_dir / hexsha
            for i in range(1000):
                if not run_fold.exists():
                    break
                run_fold = run_dir / f"{hexsha}_{i}"
            git.Repo.clone_from(self.path.as_posix(), run_fold.as_posix())
        args = copy(tao.args)
        del args.tao_dirty
        del args.tao_cmd
        del args.tao_commit
        for key, val in metadata.items():
            args.training_script_args += [f"--{key}", val]
        prev_cwd = os.getcwd()
        os.chdir(run_fold)
        os.environ[
            "PYTHONPATH"
        ] = f'{run_fold.as_posix()}:{os.getenv("PYTHONPATH", "")}'
        run(args)
        os.chdir(prev_cwd)

    @classmethod
    def create(cls, path: Union[Path, str]):
        """Create a tao project from scratch"""
        if isinstance(path, str):
            path = Path(path)
        path.mkdir(exist_ok=False)
        repo = cls(path)
        repo.init()
        return repo

    @classmethod
    def find_by_file(cls, path: Union[Path, str]):
        """Find the nearest tao repo of any file"""
        path = Path(path).resolve()
        while True:
            if path.as_posix() == "/":
                raise FileNotFoundError()
            if (path / ".tao").exists():
                break
            path = path.parent
        return cls(path)

    @tao.ensure_config("kaggle_username", "kaggle_key", "kaggle_dataset_id")
    def sync_code_to_kaggle(self):
        """Create a GitHub workflow that sync the source code to Kaggle dataset.

        There are presteps before generating this action:
        1. Create a dataset with any file(which will give the `dataset_slug` parameter)
        2. In the GitHub repo settings, add two GitHub action secrets named KAGGLE_USERNAME and KAGGLE_KEY
        """

        # kaggle do authentication when importing the package
        # so the import goes here
        os.environ["KAGGLE_USERNAME"] = tao.cfg["kaggle_username"]
        os.environ["KAGGLE_KEY"] = tao.cfg["kaggle_key"]
        import kaggle

        tempdir = Path(mkdtemp())
        output_file = tempdir / "output.zip"
        with output_file.open("wb") as f:
            self.git.archive(f, format="zip")
        metadata_file = tempdir / "dataset-metadata.json"
        metadata = """{
  "id": "%s",
  "licenses": [
    {
      "name": "CC0-1.0"
    }
  ]
}
""" % (
            tao.cfg["kaggle_dataset_id"],
        )
        metadata_file.write_text(metadata)
        kaggle.api.dataset_create_version(
            folder=tempdir, version_notes=self.git.head.ref.commit.message
        )
        shutil.rmtree(tempdir)



import argparse
import os
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, List

import jinja2

import yaml
from torch.distributed.run import get_args_parser

import pytorch_tao as tao


def load_cfg(cfg_path: Path) -> Dict:
    """Load global config, all the config can be accessed by tao.cfg

    By default, tao will load default cfg, if TAO_ENV is set, tao will load the
    specific part of config. TAO_ENV can be used to seperate different running
    environment, like local machine and colab... All the other part of config will
    inherit from default.
    """
    config_name = os.getenv("TAO_ENV", "default")
    with cfg_path.open("r") as f:
        config = yaml.load(f, Loader=yaml.Loader) or {}
    default_config = config.get("default", {})
    env_config = config.get(config_name, {})
    default_config.update(env_config)
    tao.cfg = default_config


class ConfigMissingError(Exception):
    def __init__(self, keys: List[str], func: Callable):
        self.missing_keys = set(keys)
        self.func = func
        super().__init__(
            f"Config keys {self.missing_keys} must be present for calling {func.__name__}"
        )


def ensure_config(*keys):
    """A decorator to ensure that some config keys must be present for a function"""

    def decorator(func):
        @wraps(func)
        def real(*args, **kwargs):
            missing_keys = [
                key for key in keys if key not in tao.cfg or not tao.cfg[key]
            ]
            if len(missing_keys) != 0:
                raise ConfigMissingError(missing_keys, func)
            return func(*args, **kwargs)

        return real

    return decorator


_template_renderer = jinja2.Environment(
    loader=jinja2.PackageLoader("pytorch_tao"), autoescape=jinja2.select_autoescape()
)


def render_tpl(template_name, **kwargs) -> str:
    template = _template_renderer.get_template(template_name + ".jinja")
    return template.render(**kwargs)


def parse_args(args: str = None):
    parser = argparse.ArgumentParser(description="PyTorch Tao")
    subparsers = parser.add_subparsers(dest="tao_cmd")

    run_parser = subparsers.add_parser(
        "run",
        help="Run a training process",
        parents=[get_args_parser()],
        add_help=False,
    )

    run_parser.add_argument(
        "--dirty",
        action="store_true",
        dest="tao_dirty",
        default=False,
        help="If the git repo is not clean, this option should be enabled to start the training, "
        "otherwise Tao will complain about it. Dirty run is for code testing purpose.",
    )

    run_parser.add_argument(
        "--commit",
        type=str,
        dest="tao_commit",
        default=None,
        help="Commit and run the code, it is equal to `git add -A; git commit -m xxx; tao run xxx`",
    )

    new_parser = subparsers.add_parser(
        "new",
        help="Create a tao project",
    )

    new_parser.add_argument("path", type=str, help="Path of this new project")

    tao.args = parser.parse_args(args)


def run():
    tao.repo = tao.Repo.find_by_file(tao.args.training_script)
    tao.load_cfg(tao.repo.cfg_path)
    tao.repo.run()


def new():
    tao.Repo.create(tao.args.path)


_cmd = {
    "run": run,
    "new": new,
}


def dispatch():
    if tao.args is None:
        raise RuntimeError("Should parse args before dispatch")
    _cmd[tao.args.tao_cmd]()



import os

import click


@click.group()
def cli():
    pass


@cli.command()
def format():
    os.system("poetry run black .")
    os.system("poetry run usort format .")
    os.system(
        "poetry run flake8 . --count --exit-zero --max-complexity=10 --max-line-length=127 --statistics"
    )


def main():
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytorch_tao']

package_data = \
{'': ['*'], 'pytorch_tao': ['templates/*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'click>=8.1.3,<9.0.0',
 'kaggle>=1.5.12,<2.0.0',
 'torch>=1.12.1,<2.0.0']

entry_points = \
{'console_scripts': ['tao = pytorch_tao.cli:main',
                     'tao_devtool = pytorch_tao.devtool:main']}

setup_kwargs = {
    'name': 'pytorch-tao',
    'version': '0.1.4',
    'description': 'Tao for PyTorch',
    'long_description': 'None',
    'author': 'chenglu',
    'author_email': 'chenglu.she@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
